use std::collections::{HashMap, VecDeque};
use color_eyre::Result;
use crossterm::event::{KeyCode, KeyEvent};
use ratatui::{
    layout::{Alignment, Rect}, 
    prelude::*, 
    style::Style, 
    text::Line, 
    widgets::{Block, BorderType, Paragraph},
    Frame
};
use sp_consensus_babe::digests::PreDigest;
use sp_runtime::DigestItem;
use codec::Decode;

use super::Component;
use crate::{action::Action, app::Mode, palette::StylePalette};

#[derive(Debug, Clone)]
struct BlockInfo {
    block_number: u32,
    finalized: bool,
    hash: String,
}

#[derive(Debug, Default)]
pub struct ExplorerBlocks {
    blocks: VecDeque<BlockInfo>,
    extrinsics: HashMap<String, Vec<String>>,
    logs: HashMap<String, Vec<String>>,
    validators: Vec<String>,
    palette: StylePalette,
    max_block_len: u32,

    is_active: bool,
    used_paragraph_index: usize,
    used_block_index: Option<(usize, u32)>,
    used_ext_index: Option<(String, usize, usize)>,
}

impl ExplorerBlocks {
    const MAX_BLOCKS:           usize = 50;
    const LENGTH_OF_BLOCK_HASH: u16   = 66; // hash + 0x prefix
    const LENGTH_OF_ADDRESS:    u16   = 49;
    const TOTAL_OFFSETS:        u16   = 18;

    fn update_validator_list(&mut self, validators: Vec<String>) -> Result<()> {
        self.validators = validators;
        Ok(())
    }

    fn get_author_from_digest(&self, logs: Vec<String>) -> Option<String> {
        let logs = logs
            .iter()
            .map_while(|log| {
                hex::decode(log.trim_start_matches("0x"))
                    .ok()
                    .map(|log_hex| DigestItem::decode(&mut &log_hex[..]).ok())
            })
            .filter_map(|digest| digest)
            .collect::<Vec<_>>();

        let maybe_author = match logs.iter().find(|item| matches!(item, DigestItem::PreRuntime(..))) {
            Some(DigestItem::PreRuntime(engine, data)) if *engine == [b'B', b'A', b'B', b'E'] => {
                match PreDigest::decode(&mut &data[..]) {
                    Ok(PreDigest::Primary(primary)) => self.validators.get(primary.authority_index as usize),
                    Ok(PreDigest::SecondaryPlain(secondary)) => self.validators.get(secondary.authority_index as usize),
                    Ok(PreDigest::SecondaryVRF(secondary)) => self.validators.get(secondary.authority_index as usize),
                    _ => None,
                }
            },
            _ => None,
        };

        maybe_author.cloned()
    }

    fn update_latest_block_info(
        &mut self, 
        hash: String, 
        number_hex_str: String, 
        logs: Vec<String>,
        extrinsics: Vec<String>,
    ) -> Result<()> {
        let number_hex_str = number_hex_str.trim_start_matches("0x");
        let block_number = u32::from_str_radix(&number_hex_str, 16)?;
        let front_block_number = if self.blocks.is_empty() {
            0
        } else {
            self.blocks.front().unwrap().block_number
        };

        if front_block_number < block_number {
            self.blocks.push_front(BlockInfo {
                block_number, 
                finalized: false,
                hash: hash.clone(), 
            });

            self.extrinsics.insert(hash.clone(), extrinsics);
            self.logs.insert(hash, logs);

            let block_length = block_number.checked_ilog10().unwrap_or(0) + 1;
            if self.max_block_len < block_length {
                self.max_block_len = block_length;
            }

            if self.blocks.len() > Self::MAX_BLOCKS {
                if let Some(removed_block_info) = self.blocks.pop_back() {
                    self.extrinsics.remove(&removed_block_info.hash);
                    self.logs.remove(&removed_block_info.hash);
                }
            }
        }
        Ok(())
    }

    fn update_finalized_block_info(
        &mut self, 
        hash: String, 
        number_hex_str: String, 
        logs: Vec<String>,
        extrinsics: Vec<String>,
    ) -> Result<()> {
        let number_hex_str = number_hex_str.trim_start_matches("0x");
        let block_number = u32::from_str_radix(&number_hex_str, 16)?;

        for idx in 0..self.blocks.len() {
            if self.blocks[idx].finalized { break; }
            else if self.blocks[idx].block_number > block_number { continue; }
            else {
                self.blocks[idx].finalized = true;
                self.blocks[idx].hash = hash.clone();
                self.extrinsics.insert(hash.clone(), extrinsics.clone());
                self.logs.insert(hash.clone(), logs.clone());
            }
        }

        Ok(())
    }

    fn prepare_block_line_info(&self, current_block: &BlockInfo, width: u16) -> Line {
        let block_number_length = self
            .max_block_len
            .max(current_block.block_number.checked_ilog10().unwrap_or(0) + 1) as usize;

        let free_space = width
            .saturating_sub(block_number_length as u16)
            .saturating_sub(Self::TOTAL_OFFSETS);

        let author = self
            .logs
            .get(&current_block.hash)
            .map_or(String::from("..."), |maybe_logs| {
                self.get_author_from_digest(maybe_logs.to_vec())
                    .map_or(String::from("..."), |maybe_author| maybe_author)
            });

        if free_space < Self::LENGTH_OF_BLOCK_HASH + Self::LENGTH_OF_ADDRESS {
            let len_for_author = free_space * Self::LENGTH_OF_ADDRESS / (Self::LENGTH_OF_BLOCK_HASH + Self::LENGTH_OF_ADDRESS);
            let len_for_hash = (free_space - len_for_author) / 2;

            let hash_to_print = format!("{}...{}", 
                &current_block.hash[..len_for_hash as usize],
                &current_block.hash[(Self::LENGTH_OF_BLOCK_HASH - len_for_hash) as usize..]);

            if &author == "..." {
                Line::raw(format!("{:^left$}| {} | {:^right$}", 
                        current_block.block_number, 
                        hash_to_print,
                        author,
                        left=block_number_length,
                        right=(len_for_author + 2) as usize))
            } else {
                Line::raw(format!("{} | {} | {}", 
                        current_block.block_number, 
                        hash_to_print,
                        format!("{}...", &author[..(len_for_author) as usize])))
            }
        } else {
            let total_space_used = block_number_length as u16 + Self::LENGTH_OF_BLOCK_HASH + Self::LENGTH_OF_ADDRESS;
            let margin = (width - total_space_used) as usize / 3;
            Line::raw(format!("{:^margin$}|{:^margin$}|{:^margin$}",
                    current_block.block_number, 
                    current_block.hash, 
                    author))
        }
    }

    fn prepare_block_lines(&mut self, rect: Rect) -> Vec<Line> {
        let width = rect.as_size().width;
        let total_length = rect.as_size().height as usize - 2;
        let mut items = Vec::new();

        let active_style = self.palette.create_text_style(true);
        let latest_style = self.palette.create_text_style(false);
        let finalized_style = Style::new().fg(self.palette.foreground_hover());

        let start_index = match self.used_block_index {
            Some((_, used_block)) if total_length < self.blocks.len() => {
                self.blocks
                    .iter()
                    .position(|b| b.block_number == used_block)
                    .unwrap_or_default()
                    .saturating_add(1)
                    .saturating_sub(total_length)
            },
            _ => 0,
        };

        for (idx, current_block_info) in self.blocks.iter().skip(start_index).enumerate() {
            if idx == total_length { break; }

            let style = match self.used_block_index {
                Some((_, used_block)) if current_block_info.block_number == used_block => active_style,
                _ => {
                    if current_block_info.finalized { finalized_style } else { latest_style }
                }
            };

            items.push(self.prepare_block_line_info(&current_block_info, width).style(style));
        }

        items
    }

    fn prepare_ext_line_info(&self, index: usize, extrinsic: String, width: u16) -> Line {
        let index_len = index.checked_ilog10().unwrap_or(0) + 1;
        let len_for_ext = width.saturating_sub(index_len as u16 + 17) as usize;
        let len_extrinsic_hash = extrinsic.len();
        Line::from(format!("{} MODULE METHOD {}", 
                index, 
                format!("{}...{}", &extrinsic[..len_for_ext], &extrinsic[len_extrinsic_hash - len_for_ext..]))) 
    }

    fn prepare_ext_lines(&mut self, rect: Rect) -> Vec<Line> {
        let width = rect.as_size().width;
        let mut total_length = rect.as_size().height - 2;
        let mut items = Vec::new();

        let normal_style = self.palette.create_text_style(false);
        let active_style = self.palette.create_text_style(true);

        if let Some((_, used_block_number)) = self.used_block_index {
            let hash = self.blocks
                .iter()
                .find(|b| b.block_number == used_block_number)
                .map(|b| b.hash.clone())
                .unwrap_or_default();

            if let Some(exts) = self.extrinsics.get(&hash) {
                for (index, ext) in exts.iter().enumerate() {
                    if total_length == 0 { break; }

                    let style = if let Some((_, _, used_ext_index)) = self.used_ext_index {
                        if index == used_ext_index { active_style } else { normal_style }
                    } else {
                        normal_style
                    };

                    items.push(self.prepare_ext_line_info(0, ext.to_string(), width).style(style));
                    total_length -= 1;
                }
            }

        }

        items
    }

    fn prepare_event_lines(&mut self, rect: Rect) -> Line {
        let _width = rect.as_size().width;

        let style = self.palette.create_text_style(false);

        if let Some((hash, _, ext_index)) = &self.used_ext_index {
            if let Some(exts) = &self.extrinsics.get(&hash.clone()) {
                Line::from(format!("{}", exts.get(*ext_index).unwrap_or(&String::from("nothing here")))).style(style)
            } else {
                Line::from("nothing here")
            }
        } else {
            Line::from("nothing here")
        }
    }

    fn move_right(&mut self) {
        let new_index = self.used_paragraph_index + 1;
        if new_index < 2 {
            self.used_paragraph_index = new_index;
        }
    }
    
    fn move_left(&mut self) {
        self.used_paragraph_index = self
            .used_paragraph_index
            .saturating_sub(1);
        self.used_ext_index = None;
    }

    fn move_down(&mut self) {
        if self.used_paragraph_index == 0 {
            self.move_down_blocks();
        } else {
            self.move_down_extrinsics();
        }
    }

    fn move_up(&mut self) {
        if self.used_paragraph_index == 0 {
            self.move_up_blocks();
        } else {
            self.move_up_extrinsics();
        }
    }

    fn move_up_extrinsics(&mut self) {
        match &self.used_ext_index {
            Some((header, block_index, used_index)) => {
                if *used_index == 0 { return }
                let new_index = used_index - 1;

                let maybe_exts = self.extrinsics.get(&*header);
                if maybe_exts.is_none() { return }

                let found = maybe_exts
                    .unwrap()
                    .get(new_index)
                    .is_some();

                if found {
                    self.used_ext_index = 
                        Some(((&*header).clone(), *block_index, new_index));
                }
                
            },
            None => {
                self.used_ext_index = self.blocks
                    .front()
                    .map(|block| {
                        self.extrinsics
                            .get(&block.hash)
                            .map(|_| (block.hash.clone(), 0, 0))
                    })
                .flatten()
            }
        }
    }

    fn move_up_blocks(&mut self) {
        self.used_block_index = match &self.used_block_index {
            Some((used_index, used_block)) => {
                Some(self.blocks
                    .iter()
                    .enumerate()
                    .find(|(_, b)| b.block_number == used_block + 1)
                    .map(|(idx, b)| (idx, b.block_number))
                    .unwrap_or((*used_index, *used_block)))
            },
            None => self.blocks.front().map(|b| (0usize, b.block_number)),
        }
    }

    fn move_down_extrinsics(&mut self) {
        match &self.used_ext_index {
            Some((header, block_index, used_index)) => {
                let new_index = used_index + 1;

                let maybe_exts = self.extrinsics.get(&*header);
                if maybe_exts.is_none() { return }
                let exts = maybe_exts.unwrap();

                let found = exts
                    .get(new_index)
                    .is_some();

                if found && new_index < exts.len() {
                    self.used_ext_index = 
                        Some(((&*header).clone(), *block_index, new_index));
                }
            },
            None => {
                self.used_ext_index = self.blocks
                    .front()
                    .map(|block| {
                        self.extrinsics
                            .get(&block.hash)
                            .map(|_| (block.hash.clone(), 0, 0))
                    })
                .flatten()
            }
        }
    }

    fn move_down_blocks(&mut self) {
        self.used_block_index = match &self.used_block_index {
            Some((used_index, used_block)) => {
                Some(self.blocks
                    .iter()
                    .enumerate()
                    .find(|(_, b)| b.block_number == used_block.saturating_sub(1))
                    .map(|(idx, b)| (idx, b.block_number))
                    .unwrap_or((*used_index, *used_block)))
            },
            None => {
                self.blocks.front().map(|b| (0usize, b.block_number))
            }
        }
    }

    fn set_active(&mut self) -> Result<()> {
        self.is_active = true;
        Ok(())
    }

    fn unset_active(&mut self) -> Result<()> {
        self.is_active = false;
        Ok(())
    }

    fn prepare_blocks_paragraph(
        &mut self, 
        place: Rect, 
        border_style: Color,
        border_type: BorderType,
    ) -> Paragraph {
        let title_style = self.palette.create_title_style();
        Paragraph::new(self.prepare_block_lines(place))
            .block(Block::bordered()
                .border_style(border_style)
                .border_type(border_type)
                .title_alignment(Alignment::Right)
                .title_style(title_style)
                .title("Blocks"))
            .alignment(Alignment::Center)
    }

    fn prepare_extrinsics_paragraph(
        &mut self, 
        place: Rect, 
        border_style: Color,
        border_type: BorderType,
    ) -> Paragraph {
        let title_style = self.palette.create_title_style();
        Paragraph::new(self.prepare_ext_lines(place))
            .block(Block::bordered()
                .border_style(border_style)
                .border_type(border_type)
                .title_alignment(Alignment::Right)
                .title_style(title_style)
                .title("Transactions"))
            .alignment(Alignment::Center)
    }

    fn prepare_event_paragraph(
        &mut self, 
        place: Rect, 
        border_style: Color,
        border_type: BorderType,
    ) -> Paragraph {
        let title_style = self.palette.create_title_style();
        Paragraph::new(self.prepare_event_lines(place))
            .block(Block::bordered()
                .border_style(border_style)
                .border_type(border_type)
                .title_alignment(Alignment::Right)
                .title_style(title_style)
                .title("Events"))
            .alignment(Alignment::Center)
    }
}

impl Component for ExplorerBlocks {
    fn handle_key_event(&mut self, key: KeyEvent) -> Result<Option<Action>> {
        match key.code {
            KeyCode::Char('k') | KeyCode::Up if self.is_active => self.move_up(), 
            KeyCode::Char('j') | KeyCode::Down if self.is_active => self.move_down(), 
            KeyCode::Char('l') | KeyCode::Right if self.is_active => self.move_right(), 
            KeyCode::Char('h') | KeyCode::Left if self.is_active => self.move_left(), 
            KeyCode::Esc => {
                self.used_block_index = None;
                self.used_ext_index = None;
                self.used_paragraph_index = 0;
            }, 
            _ => {},
        };

        Ok(None)
    }

    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::SetLatestBlock(hash, block) => self.update_latest_block_info(hash, block.header.number, block.header.digest.logs, block.extrinsics)?,
            Action::SetFinalizedBlock(hash, block) => self.update_finalized_block_info(hash, block.header.number, block.header.digest.logs, block.extrinsics)?,
            Action::SetValidators(validators) => self.update_validator_list(validators)?,
            Action::SetMode(Mode::ExplorerActive) if !self.is_active => self.set_active()?,
            Action::SetMode(_) if self.is_active => self.unset_active()?,
            _ => {}
        };
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [blocks_place, ext_place] = super::explorer_scrollbars_layout(area);
        let [_, _, event_place] = super::explorer_layout(area);

        let (border_style_block, border_type_block) = self.palette.create_border_style(self.is_active && self.used_paragraph_index == 0);
        let (border_style_extrinsics, border_type_extrinsics) = self.palette.create_border_style(self.is_active && self.used_paragraph_index == 1);
        // TODO: never used, revisit
        let (border_style_event, border_type_event) = self.palette.create_border_style(self.is_active && self.used_paragraph_index == 2);

        frame.render_widget(self.prepare_blocks_paragraph(blocks_place, border_style_block, border_type_block), blocks_place);
        frame.render_widget(self.prepare_extrinsics_paragraph(ext_place, border_style_extrinsics, border_type_extrinsics), ext_place);
        frame.render_widget(self.prepare_event_paragraph(event_place, border_style_event, border_type_event), event_place);
        Ok(())
    }
}
